from autosubmit.config.log import Log

from earthdiagnostics import Utils, TempFile, cdftools
from earthdiagnostics.diagnostic import Diagnostic
import shutil
import os
import numpy as np
import threading


class VerticalMean(Diagnostic):
    """
        Choose vertical level in ocean, or vertically average between
        2 or more  ocean levels

        Created in  February 2012      Author : vguemas@ic3.cat
        Modified (more generic, i.e. for any input var) in December 2014
        Author : eleftheria.exarchou@ic3.cat

        :param input_file: input file name
        :param output_file: output file name  (=> 2D field  )
        :param variable: variable name
        :param lev_min: upper depth of the layer (in level number)
        :param lev_max: lower depth of the layer (in level number. If same as the upper layer, then the script just
        selects this layer. If its different, a vertical integral (weighted) between upper and lower level is computed
        """

    def __init__(self, data_manager, startdate, member, chunk, variable, min_lev, max_lev):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.variable = variable
        self.min_lev = min_lev
        self.max_lev = max_lev
        self.required_vars = [variable]
        self.generated_vars = [self._get_output_var()]

    def _get_output_var(self):
        return '{0}mean{1}-{2}'.format(self.variable, self.min_lev, self.max_lev)

    @classmethod
    def generate_jobs(cls, data_manager, startdates, members, chunks, options):
        num_options = len(options) -1
        if num_options < 1:
            raise Exception('You must specify the variable to average vertically')
        if num_options > 3:
            raise Exception('You must specify between one and three parameters for the vertical mean')
        variable = options[1]
        if num_options >= 2:
            min_lev = int(options[2])
        else:
            min_lev = None
        if num_options >= 2:
            max_lev = int(options[3])
        else:
            max_lev = None

        job_list = list()
        for startdate in startdates:
            for member in members:
                for chunk in range(1, chunks + 1):
                    job_list.append(VerticalMean(data_manager, startdate, member, chunk,
                                                 variable, min_lev, max_lev))
        return job_list

    def compute(self):
        temp = TempFile.get()
        variable_file = self.data_manager.get_file('ocean', self.variable, self.startdate, self.member, self.chunk)

        handler = Utils.nco.openCdf(variable_file)
        if self.min_lev is None:
            lev_min = handler.variables['lev'][0]
        else:
            lev_min = handler.variables['lev'][self.min_lev]

        if self.min_lev is None:
            lev_max = handler.variables['lev'][-1]
        else:
            lev_max = handler.variables['lev'][self.max_lev]
        handler.close()

        cdftools.run('cdfvertmean', input=variable_file, output=temp, options=[self.variable, 'T', lev_min, lev_max,
                                                                               '-debug'])
        Utils.setminmax(temp, '{0}_vert_mean'.format(self.variable))
        Utils.rename_variable(temp, '{0}_vert_mean'.format(self.variable), self._get_output_var())
        self.data_manager.send_file(temp, 'ocean', self._get_output_var(), self.startdate, self.member, self.chunk)


class VerticalMeanMeters(Diagnostic):
    """
        Vertically averaged salt content
        Created in February 2012      Author : vguemas@ic3.cat

        :param input_file: input grid_T file name
        :param upper: upper depth of the layer (in meters)
        :type upper: int
        :param lower: lower depth of the layer (in meters)
        :type lower: int
        :param output_file: output file name  (=> 2D)
    """

    def __init__(self, data_manager, startdate, member, chunk, variable, min_depth, max_depth):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.variable = variable
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.required_vars = [variable]
        self.generated_vars = [self._get_output_var()]

    def _get_output_var(self):
        return '{0}mean{1}m-{2}m'.format(self.variable, self.min_depth, self.max_depth)

    @classmethod
    def generate_jobs(cls, data_manager, startdates, members, chunks, options):
        num_options = len(options) - 1
        if num_options < 1:
            raise Exception('You must specify the variable to average vertically')
        if num_options > 3:
            raise Exception('You must specify between one and three parameters for the vertical mean')
        variable = options[1]
        if num_options >= 2:
            min_lev = float(options[2])
        else:
            min_lev = None
        if num_options >= 2:
            max_lev = float(options[3])
        else:
            max_lev = None

        job_list = list()
        for startdate in startdates:
            for member in members:
                for chunk in range(1, chunks + 1):
                    job_list.append(VerticalMeanMeters(data_manager, startdate, member, chunk,
                                                       variable, min_lev, max_lev))
        return job_list

    def compute(self):
        temp = TempFile.get()
        variable_file = self.data_manager.get_file('ocean', self.variable, self.startdate, self.member, self.chunk)

        handler = Utils.nco.openCdf(variable_file)
        if self.min_depth is None:
            lev_min = handler.variables['lev'][0]
        else:
            lev_min = self.min_depth

        if self.min_depth is None:
            lev_max = handler.variables['lev'][-1]
        else:
            lev_max = self.max_depth
        handler.close()

        cdftools.run('cdfvertmean', input=variable_file, output=temp, options=[self.variable, 'T', lev_min, lev_max,
                                                                               '-debug'])
        Utils.setminmax(temp, '{0}_vert_mean'.format(self.variable))
        Utils.rename_variable(temp, '{0}_vert_mean'.format(self.variable), self._get_output_var())
        self.data_manager.send_file(temp, 'ocean', self._get_output_var(), self.startdate, self.member, self.chunk)


class General(object):
    """
    General diagnostics that can be used with different variables
    """


    @staticmethod
    def interpolate(input_file, output_file, variable, nemo_version):
        """
        3-dimensional conservative interpolation to the regular atmospheric grid

        Created in November 2012      Author : vguemas@ic3.cat

        :param nemo_version:
        :param input_file:
        :param output_file:
        :param variable:
        :return:
        """
        temp = TempFile.get()
        temp2 = TempFile.get()
        shutil.copy(input_file, temp)

        cdo = Utils.cdo
        nco = Utils.nco
        handler = cdo.openCdf(temp)
        if 'lev' in handler.dimensions:
            num_levels = handler.dimensions['lev'].size
            has_levels = True
        else:
            num_levels = 1
            has_levels = False
        handler.close()

        threads = list()
        numthreads = Utils.available_cpu_count()

        for numthread in range(0, numthreads):
            t = threading.Thread(target=General._interpolate_level,
                                 args=(range(numthread, num_levels, numthreads),has_levels, temp, variable,
                                       nemo_version))
            threads.append(t)
            t.start()

        for t in threads:
            t.join()
        if has_levels:
            nco.ncrcat(input='tmp_01.nc', output=temp2,
                       options='-n {0},2,1 -v {1}'.format(num_levels, variable))

        else:
            Utils.move_file('tmp_01.nc', temp2)

        handler = nco.openCdf(temp2)
        handler.renameDimension('record', 'lev')
        handler.close()
        nco.ncpdq(input=temp2, output=temp2, options='-O -h -a time,lev')

        if has_levels:
            nco.ncks(input=temp, output=temp2, options='-A -v lev')
            for lev in range(0, num_levels):
                os.remove('tmp_{0:02d}.nc'.format(lev + 1))

        cdo.setgrid('t106grid', input=temp2, output=temp)
        if nemo_version[6:9] == '025':
            cdo.invertlatdata(input=temp, output=temp)
        if not has_levels:
            nco.ncks(input=temp, output=temp, options='-O -v sic,lat,lon,time')

        os.remove(temp2)
        Utils.move_file(temp, output_file)

    @staticmethod
    def _interpolate_level(levs, has_levels, input_file, variable, nemo_version):
        for lev in levs:
            nco = Utils.nco
            temp = TempFile.get()
            if has_levels:
                nco.ncks(input=input_file, output=temp, options='-O -d lev,{0} -v {1}'.format(lev, variable))
                nco.ncwa(input=temp, output=temp, options='-O -h -a lev')
            else:
                shutil.copy(input_file, temp)
            namelist_file = 'scrip_use_in{0}'.format(lev)
            scrip_use_in = open(namelist_file, 'w')
            scrip_use_in.writelines("&remap_inputs\n")
            scrip_use_in.writelines("   remap_wgt   = '/esnas/autosubmit/con_files/"
                                    "weigths/{0}/rmp_{0}_to_regular_lev{1}.nc'\n".format(nemo_version, lev+1))
            scrip_use_in.writelines("   infile      = '{0}'\n".format(temp))
            scrip_use_in.writelines("   invertlat   = FALSE\n")
            scrip_use_in.writelines("   var         = '{0}'\n".format(variable))
            scrip_use_in.writelines("   fromregular = FALSE\n")
            scrip_use_in.writelines("   outfile     = '{0}'\n".format(temp))
            scrip_use_in.writelines("/\n")
            scrip_use_in.close()
            Utils.execute_shell_command('/home/Earth/jvegas/pyCharm/cfutools/interpolation/scrip_use '
                                        '{0}'.format(namelist_file), Log.NO_LOG)
            os.remove(namelist_file)
            nco.ncecat(input=temp, output=temp, options="-O -h")
            shutil.move(temp, 'tmp_{0:02d}.nc'.format(lev + 1))
            Log.debug("Level {0} ready", lev)

    @staticmethod
    def cut_section(input_file, output_file, variable, zonal, value):
        """
        Cut a meridional or zonal section

        Created in September 2012     Author : vguemas@ic3.cat
        :param input_file: input file
        :param output_file: output file  ( => 2D )
        :param variable: input var
        :param zonal: (zonal / meridional section)
        :param value: lat/lon
        :return:
        """

        nco = Utils.nco

        handler = nco.openCdf('mesh_hgr.nc')
        dimi = handler.dimensions['i'].size
        dimj = handler.dimensions['j'].size
        dimlev = handler.dimensions['lev'].size

        lon = handler.variables['lon'][:]
        lat = handler.variables['lat'][:]
        handler.close()

        handler = nco.openCdf('mask.nc')
        mask_lev = handler.variables['tmask'][:]
        mask_lev = mask_lev.astype(float)
        # noinspection PyTypeChecker
        np.place(mask_lev, mask_lev == 0, [1e20])
        handler.close()

        # Latitude / longitude of the zonal / meridional section
        exactpos = value
        if not zonal:
            while exactpos < np.min(lon):
                exactpos += 360
            while exactpos > np.max(lon):
                exactpos -= 360
            size = dimj
        else:
            size = dimi

        # Collect the indexes defining the section

        listi = np.empty(size, dtype=int)
        listj = np.empty(size, dtype=int)

        for jpt in range(0, size):
            if not zonal:
                vector = lon[jpt, :]
            else:
                vector = lat[:, jpt]
            distance = abs(vector - exactpos)
            pos = np.where(distance == min(distance))
            if not zonal:
                listi[jpt] = pos[0][0]
                listj[jpt] = jpt
            else:
                listi[jpt] = jpt
                listj[jpt] = pos[0][0]

        temp = TempFile.get()
        shutil.copy(input_file, temp)

        handler = nco.openCdf(temp)
        dimtime = handler.dimensions['time'].size
        var_array = handler.variables[variable][:]
        handler.close()

        var = np.empty([dimtime, dimlev, size], dtype=handler.variables[variable].dtype)
        new_coord = np.empty(size, dtype=float)
        if zonal:
            old_coord = lon
        else:
            old_coord = lat

        for jpt in range(0, size):
            Log.info(str(jpt))
            var[:, :, jpt] = np.maximum(var_array[:, :, listj[jpt], listi[jpt]],
                                        mask_lev[:, :, listj[jpt], listi[jpt]])
            new_coord[jpt] = old_coord[listj[jpt], listi[jpt]]

        nco.ncks(input=temp, output=temp, options='-O -v lev,time')

        handler = nco.openCdf(temp)
        if not zonal:
            handler.createDimension('lat', size)
            coord_var = handler.createVariable('lat', float, 'lat')
            file_var = handler.createVariable(variable, float, ('time', 'lev', 'lat'))
        else:
            handler.createDimension('lon', size)
            coord_var = handler.createVariable('lon', float, 'lon')
            file_var = handler.createVariable(variable, float, ('time', 'lev', 'lon'))
        coord_var[:] = new_coord[:]
        file_var[:] = var[:]
        file_var.missing_value = 1e20
        handler.close()
        Utils.move_file(temp, output_file)

    @staticmethod
    def avgsection(input_file, output_file, lon_min, lon_max, lat_min, lat_max):
        temp = TempFile.get()
        Utils.cdo.zonmean(input='-sellonlatbox,{0},{1},{2},{3} {4}'.format(lon_min, lon_max, lat_min, lat_max,
                                                                           input_file),
                          output=temp)
        Utils.move_file(temp, output_file)


